﻿import com.xfactorstudio.xml.xpath.Token;
import com.xfactorstudio.xml.xpath.TokenTypes;

class com.xfactorstudio.xml.xpath.Tokenizer{
    private var xpath:String;
    private var currentPosition:Number;
    private var endPosition:Number;

    private var previousToken:Token = null;

    public function Tokenizer(xpath:String){
        setXPath( xpath );
    }

    public function setXPath(xpath:String):Void {
        this.xpath           = xpath;
        this.currentPosition = 0;
        this.endPosition     = xpath.length;
    }

    public function nextToken():Token {
        var token:Token = null;
        do{
            token = null;
            switch ( LA(1) ){
                case '$':
                    token = dollar();
                    break;
                case '"':
                case '\'':
                    token = literal();
                    break;
                case '/':
                    token = slashes();
                    break;
                case ',':
                    token = comma();
                    break;
                case '(':
                    token = leftParen();
                    break;
                case ')':
                    token = rightParen();
                    break;
                case '[':
                    token = leftBracket();
                    break;
                case ']':
                    token = rightBracket();
                    break;
                case '+':
                    token = plus();
                    break;
                case '-':
                    token = minus();
                    break;
                case '<':
                case '>':
                    token = relationalOperator();
                    break;
               	case '=':
                    token = equals();
                    break;
                case '!':
                    if ( LA(2) == '=' ){
                        token = notEquals();
                    }else{
                        token = Not();
                    }
                    break;
                case '|':
                    token = pipe();
                    break;
                case '@':
                    token = at();
                    break;
                case ':':
                    if ( LA(2) == ':' ){
                        token = doubleColon();
                    }else{
                        token = colon();
                    }
                    break;
                case '*':
					token = star();
					break;
               	case '.':
                    switch ( LA(2) ){
                        case '0':
                        case '1':
                        case '2':
                        case '3':
                        case '4':
                        case '5':
                        case '6':
                        case '7':
                        case '8':
                        case '9':
                            token = numberTok();
                            break;
                        default:
                            token = dots();
                            break;
                    }
                    break;
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    token = numberTok();
                    break;
                case ' ':
                case '\t':
                case '\n':
                case '\r':
                    token = whitespace();
                    break;
                default:
                    if ( isIdentifierStartChar( LA(1) ) ){
                        token = identifierOrOperatorName();
                    }
            }

            if ( token == null ) {
                token = new Token(TokenTypes.EOF, xpath.substring(currentPosition,endPosition));
            }
        }while ( token.getTokenType() == TokenTypes.SKIP);
        
        this.previousToken = token;
        
        return token;
    }

	public function identifierOrOperatorName():Token {
	    var token:Token = null;
	    if ( this.previousToken != null ){
	        // For some reason, section 3.7, Lexical structure,
	        // doesn't seem to feel like it needs to mention the
	        // SLASH, DOUBLE_SLASH, and COLON tokens for the test
	        // if an NCName is an operator or not.
	        //
	        // Accoring to section 3.7, "/foo" should be considered
	        // as a SLASH following by an OperatorName being 'foo'.
	        // Which is just simply, clearly, wrong, in my mind.
	        //
	        //     -bob
	        
	        switch ( this.previousToken.getTokenType() ){
	            case TokenTypes.AT:
	            case TokenTypes.DOUBLE_COLON:
	            case TokenTypes.LEFT_PAREN:
	            case TokenTypes.LEFT_BRACKET:
	            case TokenTypes.AND:
	            case TokenTypes.OR:
	            case TokenTypes.MOD:
	            case TokenTypes.DIV:
	            case TokenTypes.COLON:
	            case TokenTypes.SLASH:
	            case TokenTypes.DOUBLE_SLASH:
	            case TokenTypes.PIPE:
	            case TokenTypes.DOLLAR:
	            case TokenTypes.PLUS:
	            case TokenTypes.MINUS:
	            case TokenTypes.STAR:
	            case TokenTypes.COMMA:
	            case TokenTypes.LESS_THAN:
	            case TokenTypes.GREATER_THAN:
	            case TokenTypes.LESS_THAN_EQUALS:
	            case TokenTypes.GREATER_THAN_EQUALS:
	            case TokenTypes.EQUALS:
	            case TokenTypes.NOT_EQUALS:
	            	
	                token = identifier();
	                break;
	            default:
	                token = operatorName();
	                break;
	        }
	    }else{
	        token = identifier();
	    }
	
	    return token;
	}

	public function identifier():Token {
	    var token:Token = null;
	
	    var start:Number = this.currentPosition;
	
	    while ( this.hasMoreChars() ){
	        if ( isIdentifierChar( LA(1) ) ){
	            consume();
	        }else{
	            break;
	        }
	    }
	    token = new Token(TokenTypes.IDENTIFIER, xpath.substring(start,currentPosition));
	
	    return token;
	}

	public function operatorName():Token {	
	    var token:Token = null;
	
	    switch (LA(1)){
	        case 'a':
	            token = And();
	            break;
	        case 'o':
	            token = Or();
	            break;
	        case 'm':
	            token = mod();
	            break;
	        case 'd':
	            token = div();
	            break;
	    }
	
	    return token;
	}

	public function mod():Token {
	    var token:Token = null;
	       token = new Token(TokenTypes.MOD, "mod");
	
	        consume();
	        consume();
	        consume();
	    return token;
	}

	public function div():Token {
	    var token:Token = null;
	        token = new Token(TokenTypes.DIV, xpath.substring(currentPosition,currentPosition+3));
	
	        consume();
	        consume();
	        consume();
	    return token;
	}

	public function And():Token {
	    var token:Token = null;
	
	    //if ( ( LA(1) == 'a' )
	    //     &&
	    //     ( LA(2) == 'n' )
	    //     &&
	    //     ( LA(3) == 'd' )
	    //     &&
	    //     ( ! isIdentifierChar( LA(4) ) ) )
	    //{
	        token = new Token(TokenTypes.AND, "and");
	        consume();
	        consume();
	        consume();
	   // }
	
	    return token;
	}

	public function Or():Token {
	    var token:Token = null;
	        token = new Token(TokenTypes.OR, "or");
	
	        consume();
	        consume();
	    return token;
	}

	public function numberTok():Token {
	    var start:Number = currentPosition;
	    var periodAllowed:Boolean = true;
	    var keepGoing:Boolean = true;
	
	    while( keepGoing ){
	        switch ( LA(1) ){
	            case '.':
	                if ( periodAllowed ){
	                    periodAllowed = false;
	                    consume();
	                }else{
	                    break;
	                }
	                break;
	            case '0':
	            case '1':
	            case '2':
	            case '3':
	            case '4':
	            case '5':
	            case '6':
	            case '7':
	            case '8':
	            case '9':
	                consume();
	                break;
	            default:
	            	keepGoing = false;
	                break;
	        }
	    }
	
	    var token:Token = null;
	
	    if ( periodAllowed )
	    {
	        token = new Token(TokenTypes.INTEGER, xpath.substring(start,currentPosition));
	    }
	    else
	    {
	        token = new Token(TokenTypes.DOUBLE, xpath.substring(start,currentPosition));
	    }
	 
	    return token;
	}

	public function whitespace():Token {
	    consume();
	    while(hasMoreChars()){
	        switch ( LA(1) ){
	            case ' ':
	            case '\t':
	            case '\n':
	            case '\r':
	                consume();
	                break;
	            default:
				    return  new Token(TokenTypes.SKIP, xpath.substring(0,0));
	        }
	    }
	}

	public function comma():Token {
	    var token:Token =  new Token(TokenTypes.COMMA, ",");
	    consume();
	
	    return token;
	}

	public function equals():Token {
	    var token:Token =  new Token(TokenTypes.EQUALS, "=");
	    consume();
	
	    return token;
	}

	public function minus():Token {
	    var token:Token = new Token(TokenTypes.MINUS, "-");
	    consume();
	        
	    return token;
	}

	public function plus():Token {
	    var token:Token = new Token(TokenTypes.PLUS, "+");
	    consume();
	
	    return token;
	}

	public function dollar():Token {
	    var token:Token = new Token(TokenTypes.DOLLAR, "$");
	    consume();
	
	    return token;
	}

	public function pipe():Token {
	    var token:Token = new Token(TokenTypes.PIPE, "|");
	    consume();
	
	    return token;
	}

	public function at():Token {
	    var token:Token = new Token(TokenTypes.AT, "@");
	    consume();
	
	    return token;
	}

	public function colon():Token {
	    var token:Token = new Token(TokenTypes.COLON, ":");
	    consume();
	
	    return token;
	}

	public function doubleColon():Token {
	    var token:Token = new Token(TokenTypes.DOUBLE_COLON, "::");
	    consume();
	    consume();
	
	    return token;
	}

	public function Not():Token {
	    var token:Token = new Token(TokenTypes.NOT, "!");
	    consume();
	
	    return token;
	}

	public function notEquals():Token {
	    var token:Token = new Token(TokenTypes.NOT_EQUALS, "!=");
	    consume();
	    consume();
	
	    return token;
	}

	public function relationalOperator():Token {
	    var token:Token = null;
	
	    switch ( LA(1) ) {
	        case '<':
	            if ( LA(2) == '=' )
	            {
	                token = new Token(TokenTypes.LESS_THAN_EQUALS, "<=");
	                consume();
	            }
	            else
	            {
	                token = new Token(TokenTypes.LESS_THAN, "<");
	            }
	
	            consume();
	            break;
	        case '>':
	            if ( LA(2) == '=' )
	            {
	                token = new Token(TokenTypes.GREATER_THAN_EQUALS, ">=");
	                consume();
	            }
	            else
	            {
	                token = new Token(TokenTypes.GREATER_THAN, ">");
	            }
	
	            consume();
	            break;
	    }
	
	    return token;
	            
	}

	public function star():Token {
	    var token:Token = new Token(TokenTypes.STAR, "*");
	    consume();
	    return token;
	}

	public function literal():Token {
	    var token:Token = null;
	    var match:String = LA(1);
	    consume();
	    var start:Number = currentPosition;
	    
	    while ( ( token == null ) && hasMoreChars() ){
	        if ( LA(1) == match ){
	            token = new Token(TokenTypes.LITERAL, xpath.substring(start,currentPosition));
	        }
	        consume();
	    }
	
	    return token;
	}
	
	public function dots():Token {
	    var token:Token = null;

	    switch ( LA(2) ){
	        case '.':
	            token = new Token(TokenTypes.DOT_DOT, "..");
	            consume();
	            consume();
	            break;
	        default:
	            token = new Token(TokenTypes.DOT, ".");
	            consume();
	            break;
	    }
	    return token;
	}

	public function leftBracket():Token {
	    var token:Token = new Token(TokenTypes.LEFT_BRACKET, "[");
	    consume();
	    return token;
	}

	public function rightBracket():Token {
	    var token:Token = new Token(TokenTypes.RIGHT_BRACKET, "]");
	    consume();
	    return token;
	}

	public function leftParen():Token {
	    var token:Token = new Token(TokenTypes.LEFT_PAREN, "(");
	    consume();
	    return token;
	}

	public function rightParen():Token {
	    var token:Token = new Token(TokenTypes.RIGHT_PAREN, ")");
	    consume();
	    return token;
	}

	public function slashes():Token {
	    var token:Token = null;
	    switch ( LA(2) ){
	        case '/':
	            token = new Token(TokenTypes.DOUBLE_SLASH, "//");
	            consume();
	            consume();
	            break;
	        default:
	            token = new Token(TokenTypes.SLASH, "/");
	            consume();
	    }
	
	    return token;
	}

	public function LA(i:Number):String {
	    //if ( this.currentPosition + ( i - 1 ) >= this.endPosition ){
	    //    return -1;
	    //}
	    return this.xpath.charAt( this.currentPosition + (i - 1) );
	}

	public function consume():Void {
		++this.currentPosition;
	}

	public function hasMoreChars():Boolean {
	    return (this.currentPosition < this.endPosition);
	}

	public function isIdentifierChar(c:String):Boolean {
	    //switch ( c ){
	    //    case '-':
	    //    case '.':
	    //        return true;
	    //}
	    return Tokenizer.isUnicodeIdentifierPart( c );
	}

	public function  isIdentifierStartChar(c:String):Boolean {
	    return ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_".indexOf(c) != -1);
	}

	public static function isUnicodeIdentifierPart(c:String):Boolean {
		return ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_.".indexOf(c) != -1);
	}

	//public static function isLetter(c:String){
	//	return ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz".indexOf(c) != -1);
	//}

	public static function isNumber(c:String):Boolean {
		return !isNaN(c);
	}

	//public static function isNonLetterIdentifierPart(c:String){
	//	return ("_-".indexOf(c) != -1);
	//
}