﻿/**
   XPathLexer
   
   @class  
   @package
   @author neeld
   @codehint 
   @example 
   @tooltip 
*/
// interfaSys QA: OK
import com.xfactorstudio.xml.xpath.Axes;
import com.xfactorstudio.xml.xpath.XPathAxisNames;
import com.xfactorstudio.xml.xpath.XPathFunctions;
import com.xfactorstudio.xml.xpath.XPathPredicateOperator;
import com.xfactorstudio.xml.xpath.Tokenizer;
import com.xfactorstudio.xml.xpath.TokenTypes;
import com.xfactorstudio.xml.xpath.Token;

class com.xfactorstudio.xml.xpath.XPathLexer {
	public var path:String;
	public var currPos:Number;
	public var buff:String;
	public var lastTokenType:Number = -1;
	public var lastToken:Token;
	public var inFunction:Number = 0;
	public var inGroup:Number = 0;
	public var inPredicate:Number = 0;
	public var handler:Object;
	public var tokenizer:Tokenizer;
	
	public static var AXIS:Number = 0;
	public static var PREDICATESTART:Number = 1;
	public static var PREDICATEEND:Number = 2;
	public static var IDENTIFIER:Number = 3;
	public static var FUNCTIONSTART:Number = 4;
	public static var FUNCTIONEND:Number = 5;
	public static var OPERATOR:Number = 6;
	public static var COMMA:Number = 7;
	public static var GROUPSTART:Number	= 8;
	public static var GROUPEND:Number = 9;
	public static var NUMBER:Number = 10;
	public static var LITTERAL:Number = 11;
	
	private var c:String;
	private var lastOpenPeran:Number;
	private var functionNames:Object;
	private var axisNames:XPathAxisNames;
	private var operatorNames:XPathPredicateOperator;
	
	public function XPathLexer(path:String){
		functionNames = XPathFunctions.Tokens;
		axisNames = new XPathAxisNames();
		operatorNames = new XPathPredicateOperator();
	}
	
	public function parse(path:String):Void {
		this.path = path;
		tokenizer = new Tokenizer(this.path);
		var t:Token;
		while(tokenizer.hasMoreChars()){
			t = tokenizer.nextToken();
			switch(t.getTokenType()){
				case TokenTypes.SLASH:
	    			onSlash(t);
	    			break;
	    		case TokenTypes.IDENTIFIER:
	    			onIdentifier(t);
	    			break;
	    		case TokenTypes.AT:
	    			onAt();
	    			break;
	    		case TokenTypes.AND:
	    		case TokenTypes.DIV:
				case TokenTypes.EQUALS:
				case TokenTypes.GREATER_THAN:
				case TokenTypes.GREATER_THAN_EQUALS:
				case TokenTypes.LESS_THAN:
				case TokenTypes.LESS_THAN_EQUALS:
				case TokenTypes.MINUS:
				case TokenTypes.MOD:
				case TokenTypes.NOT:
				case TokenTypes.NOT_EQUALS:
				case TokenTypes.OR:
				case TokenTypes.PIPE:
				case TokenTypes.PLUS:
					onOperator(t);
	    			break;
	    		case TokenTypes.INTEGER:
	    		case TokenTypes.DOUBLE:
	    			onNumber(t);
	    			break;
	    		case TokenTypes.DOT:
	    			onDot();
	    			break;
	    		case TokenTypes.DOT_DOT:
	    			onDoubleDot();
	    			break;
	    		case TokenTypes.COLON:
	    		case TokenTypes.DOUBLE_COLON:
	    			break;
	    		case TokenTypes.DOUBLE_SLASH:
	    			onDoubleSlash(t);
	    			break;
	    		case TokenTypes.EOF:
	    			break;
	    		case TokenTypes.LEFT_BRACKET:
	    			onLeftBracket(t);
	    			break;
	    		case TokenTypes.LEFT_PAREN:
	    			onLeftParen();
	    			break;
	    		case TokenTypes.LITERAL:
	    			onLitteral(t);
	    			break;
	    		case TokenTypes.RIGHT_BRACKET:
	    			onRightBracket();
	    			break;
	    		case TokenTypes.RIGHT_PAREN:
	    			onRightParen();
	    			break;
	    		case TokenTypes.STAR:
					if(tokenizer.LA(1)==":"){
						//This is a namespace wildcard
						var u:Token = new Token(TokenTypes.IDENTIFIER, t.getTokenText() + tokenizer.nextToken().getTokenText());
						//t.setTokenType() = TokenTypes.IDENTIFIER;
						// Run it twice?
						//t.setTokenText() += tokenizer.nextToken().getTokenText();
						//t.setTokenText() += tokenizer.nextToken().getTokenText();
						onIdentifier(u);
						break;
					}
	    			onStar();
	    			break;
	    		case TokenTypes.DOLLAR: //not implemented
	    		case TokenTypes.COMMA: //not implemented
	    		case TokenTypes.SKIP: //not implemented
	    		
	    			break;
	    	}
    		this.lastToken = t;
		}
	}

	public function onSlash(t:Token):Void {
		switch(this.lastTokenType){
			case XPathLexer.IDENTIFIER:
			case XPathLexer.PREDICATEEND:
			case XPathLexer.AXIS:
				this.lastTokenType = XPathLexer.OPERATOR;
				break;
		
			default:
				handler.onAxis(Axes.ROOT);
				this.lastTokenType = XPathLexer.AXIS;
				break;
		}
	}

	public function onStar():Void {
		switch(this.lastToken.getTokenType()){
			case TokenTypes.AT:
			case TokenTypes.DOUBLE_COLON:
			case TokenTypes.LEFT_PAREN:
			case TokenTypes.LEFT_BRACKET:
			case TokenTypes.AND:
	   		case TokenTypes.DIV:
			case TokenTypes.EQUALS:
			case TokenTypes.GREATER_THAN:
			case TokenTypes.GREATER_THAN_EQUALS:
			case TokenTypes.LESS_THAN:
			case TokenTypes.LESS_THAN_EQUALS:
			case TokenTypes.MINUS:
			case TokenTypes.MOD:
			case TokenTypes.NOT:
			case TokenTypes.NOT_EQUALS:
			case TokenTypes.OR:
			case TokenTypes.PIPE:
			case TokenTypes.PLUS:
				handler.onOperator("*");
				break;
			default:
				handler.onIdentifier("*");
				break;
			
		}
	}

	public function onAt():Void {
		handler.onAxis(Axes.ATTRIBUTE);
		this.lastTokenType = XPathLexer.AXIS;
	}

	public function onDot():Void {
		handler.onAxis(Axes.SELF);
		this.lastTokenType = XPathLexer.AXIS;
	}

	public function onDoubleDot():Void {
		handler.onAxis(Axes.PARENT);
		this.lastTokenType = XPathLexer.AXIS;
	}

	public function onDoubleSlash(t:Token):Void {
		handler.onAxis(Axes.DECENDANT_OR_SELF);
		this.lastTokenType = XPathLexer.AXIS;
	}

	public function onNumber(t:Token):Void {
		handler.onNumber(Number(t.getTokenText()));
		this.lastTokenType = XPathLexer.NUMBER;
	}

	public function onIdentifier(t:Token):Void {
		var tokenText:String = t.getTokenText();
		 if (tokenizer.LA(1) == '(' ){ // node/text identifier | function
		 	switch(tokenText){
		 		case "node":
		 		case "text":
			 		handler.onIdentifier(tokenText+"()");
				 	this.lastTokenType = XPathLexer.IDENTIFIER;
				 	tokenizer.nextToken();
				 	tokenizer.nextToken();
			 		return;
			 		break;
		 		default:
			 		handler.onFunctionStart(functionNames[tokenText]);
				 	this.lastTokenType = XPathLexer.FUNCTIONSTART;
				 	this.lastOpenPeran = XPathLexer.FUNCTIONSTART;
				 	tokenizer.nextToken();
		 		return;
		 		break;
		 	}
	    }else if(tokenizer.LA(1) == ':' ){
	    	if (tokenizer.LA(2) == ':' ){//AXIS
	    		tokenText += tokenizer.nextToken().getTokenText();
	    		handler.onAxis(axisNames[tokenText]);
				this.lastTokenType = XPathLexer.AXIS;
				return;
	    	}else{//QName
	    		tokenText += tokenizer.nextToken().getTokenText();
	    		tokenText += tokenizer.nextToken().getTokenText();
	    	}
	    }
	    
		switch(this.lastTokenType){
			case XPathLexer.AXIS:
			case XPathLexer.PREDICATESTART:
			case XPathLexer.GROUPSTART:
			case XPathLexer.FUNCTIONSTART:
			case XPathLexer.OPERATOR:
			case -1:
				handler.onIdentifier(tokenText);
				this.lastTokenType = XPathLexer.IDENTIFIER;
				break;
			default:
				break;
		}
	}

	public function onRightParen():Void {
		switch(this.lastOpenPeran){
			case XPathLexer.GROUPSTART:
				handler.onGroupEnd();
				this.lastTokenType = XPathLexer.GROUPEND;
				break;		
			case XPathLexer.FUNCTIONSTART:
				handler.onFunctionEnd();
				this.lastTokenType = XPathLexer.FUNCTIONEND;
				break;
		}
	}
	public function onLeftParen():Void {
		handler.onGroupStart();
		this.lastTokenType = XPathLexer.GROUPSTART;
		this.lastOpenPeran = XPathLexer.GROUPSTART;
	}

	public function onLitteral(t:Token):Void {
		handler.onLitteral(t.getTokenText());
		this.lastTokenType = XPathLexer.LITTERAL;
	}

	public function onLeftBracket(t:Token):Void {
		handler.onPredicateStart();
		this.lastTokenType = XPathLexer.PREDICATESTART;
	}

	public function onRightBracket(t:Token):Void {
		handler.onPredicateEnd();
		this.lastTokenType = XPathLexer.PREDICATEEND;
	}

	public function onOperator(t:Token):Void {
		handler.onOperator(operatorNames[t.getTokenText()]);
		this.lastTokenType = XPathLexer.OPERATOR;
	}
}